# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage,
# (C) 2020 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime, timedelta, timezone
from unittest import TestCase

from minio import xml
from minio.commonconfig import COMPLIANCE, GOVERNANCE
from minio.retention import Retention
from minio.time import utcnow


class RetentionTest(TestCase):
    def test_config(self):
        config = Retention(GOVERNANCE, utcnow() + timedelta(days=10))
        xml.marshal(config)

        config = xml.unmarshal(
            Retention,
            """<Retention xmlns="http://s3.amazonaws.com/doc/2006-03-01/">
   <Mode>COMPLIANCE</Mode>
   <RetainUntilDate>2020-10-02T00:00:00Z</RetainUntilDate>
</Retention>""",
        )
        xml.marshal(config)
        self.assertEqual(config.mode, COMPLIANCE)
        self.assertEqual(
            config.retain_until_date,
            datetime(2020, 10, 2, 0, 0, 0, 0, timezone.utc),
        )



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage,
# (C) 2015, 2016, 2017 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
minio - MinIO Python SDK for Amazon S3 Compatible Cloud Storage

    >>> from minio import Minio
    >>> client = Minio(
    ...     "play.min.io",
    ...     access_key="Q3AM3UQ867SPQQA43P2F",
    ...     secret_key="zuf+tfteSlswRu7BJ86wekitnifILbZam1KYY3TG",
    ... )
    >>> buckets = client.list_buckets()
    >>> for bucket in buckets:
    ...     print(bucket.name, bucket.creation_date)

:copyright: (C) 2015-2020 MinIO, Inc.
:license: Apache 2.0, see LICENSE for more details.
"""

__title__ = "minio-py"
__author__ = "MinIO, Inc."
__version__ = "7.2.3"
__license__ = "Apache 2.0"
__copyright__ = "Copyright 2015, 2016, 2017, 2018, 2019, 2020 MinIO, Inc."

# pylint: disable=unused-import,useless-import-alias
from .api import Minio as Minio
from .error import InvalidResponseError as InvalidResponseError
from .error import S3Error as S3Error
from .error import ServerError as ServerError
from .minioadmin import MinioAdmin as MinioAdmin



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage,
# (C) 2015 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple
from unittest import TestCase
from urllib.parse import urlunsplit

from minio.helpers import BaseURL


def generate_error(code, message, request_id, host_id,
                   resource, bucket_name, object_name):
    return '''
    <Error>
      <Code>{0}</Code>
      <Message>{1}</Message>
      <RequestId>{2}</RequestId>
      <HostId>{3}</HostId>
      <Resource>{4}</Resource>
      <BucketName>{5}</BucketName>
      <Key>{6}</Key>
    </Error>
    '''.format(code, message, request_id, host_id,
               resource, bucket_name, object_name)


class BaseURLTests(TestCase):
    def test_aws_new_baseurl_error(self):
        cases = [
            # invalid Amazon AWS host error
            "https://z3.amazonaws.com",
            "https://1234567890.s3.amazonaws.com",
            "https://1234567890.s3-accelerate.amazonaws.com",
            "https://1234567890.abcdefgh.s3-control.amazonaws.com",
            "https://s3fips.amazonaws.com",
            "https://s3-fips.s3.amazonaws.com",
            "https://s3-fips.s3accelerate.amazonaws.com",
            "https://s3-fips.s3-accelerate.amazonaws.com",
            "https://bucket.vpce.s3.us-east-1.vpce.amazonaws.com",
            "https://bucket.bucket.vpce-1a2b3c4d-5e6f.s3.us-east-1."
            "vpce.amazonaws.com",
            "https://accesspoint.accesspoint.vpce-1a2b3c4d-5e6f.s3.us-east-1."
            "vpce.amazonaws.com",
            "https://accesspoint.vpce-1123.vpce-xyz.s3.amazonaws.com",
            # use HTTPS scheme for host error
            "http://s3-accesspoint.amazonaws.com",
            # region missing in Amazon S3 China endpoint error
            "https://s3.amazonaws.com.cn",
        ]
        for endpoint in cases:
            self.assertRaises(ValueError, BaseURL, endpoint, None)

    def test_aws_new_baseurl(self):
        Case = namedtuple("Case", ["args", "result"])
        cases = [
            Case(
                ("https://s3.amazonaws.com", None),
                {
                    "s3_prefix": "s3.",
                    "domain_suffix": "amazonaws.com",
                    "region": None,
                    "dualstack": False,
                },
            ),
            Case(
                ("https://s3.amazonaws.com", "ap-south-1a"),
                {
                    "s3_prefix": "s3.",
                    "domain_suffix": "amazonaws.com",
                    "region": "ap-south-1a",
                    "dualstack": False,
                },
            ),
            Case(
                ("https://s3.us-gov-east-1.amazonaws.com", None),
                {
                    "s3_prefix": "s3.",
                    "domain_suffix": "amazonaws.com",
                    "region": "us-gov-east-1",
                    "dualstack": False,
                },
            ),
            Case(
                ("https://s3.me-south-1.amazonaws.com", "cn-northwest-1"),
                {
                    "s3_prefix": "s3.",
                    "domain_suffix": "amazonaws.com",
                    "region": "cn-northwest-1",
                    "dualstack": False,
                },
            ),
            ###
            Case(
                ("https://s3.dualstack.amazonaws.com", None),
                {
                    "s3_prefix": "s3.",
                    "domain_suffix": "amazonaws.com",
                    "region": None,
                    "dualstack": True,
                },
            ),
            Case(
                ("https://s3.dualstack.amazonaws.com", "ap-south-1a"),
                {
                    "s3_prefix": "s3.",
                    "domain_suffix": "amazonaws.com",
                    "region": "ap-south-1a",
                    "dualstack": True,
                },
            ),
            Case(
                ("https://s3.dualstack.us-gov-east-1.amazonaws.com", None),
                {
                    "s3_prefix": "s3.",
                    "domain_suffix": "amazonaws.com",
                    "region": "us-gov-east-1",
                    "dualstack": True,
                },
            ),
            Case(
                ("https://s3.dualstack.me-south-1.amazonaws.com",
                 "cn-northwest-1"),
                {
                    "s3_prefix": "s3.",
                    "domain_suffix": "amazonaws.com",
                    "region": "cn-northwest-1",
                    "dualstack": True,
                },
            ),
            ###
            Case(
                ("https://s3-accelerate.amazonaws.com", None),
                {
                    "s3_prefix": "s3-accelerate.",
                    "domain_suffix": "amazonaws.com",
                    "region": None,
                    "dualstack": False,
                },
            ),
            Case(
                ("https://s3-accelerate.amazonaws.com", "ap-south-1a"),
                {
                    "s3_prefix": "s3-accelerate.",
                    "domain_suffix": "amazonaws.com",
                    "region": "ap-south-1a",
                    "dualstack": False,
                },
            ),
            Case(
                ("https://s3-accelerate.us-gov-east-1.amazonaws.com", None),
                {
                    "s3_prefix": "s3-accelerate.",
                    "domain_suffix": "amazonaws.com",
                    "region": "us-gov-east-1",
                    "dualstack": False,
                },
            ),
            Case(
                ("https://s3-accelerate.me-south-1.amazonaws.com",
                 "cn-northwest-1"),
                {
                    "s3_prefix": "s3-accelerate.",
                    "domain_suffix": "amazonaws.com",
                    "region": "cn-northwest-1",
                    "dualstack": False,
                },
            ),
            ###
            Case(
                ("https://s3-accelerate.dualstack.amazonaws.com", None),
                {
                    "s3_prefix": "s3-accelerate.",
                    "domain_suffix": "amazonaws.com",
                    "region": None,
                    "dualstack": True,
                },
            ),
            Case(
                ("https://s3-accelerate.dualstack.amazonaws.com",
                 "ap-south-1a"),
                {
                    "s3_prefix": "s3-accelerate.",
                    "domain_suffix": "amazonaws.com",
                    "region": "ap-south-1a",
                    "dualstack": True,
                },
            ),
            Case(
                ("https://s3-accelerate.dualstack.us-gov-east-1.amazonaws.com",
                 None),
                {
                    "s3_prefix": "s3-accelerate.",
                    "domain_suffix": "amazonaws.com",
                    "region": "us-gov-east-1",
                    "dualstack": True,
                },
            ),
            Case(
                ("https://s3-accelerate.dualstack.me-south-1.amazonaws.com",
                 "cn-northwest-1"),
                {
                    "s3_prefix": "s3-accelerate.",
                    "domain_suffix": "amazonaws.com",
                    "region": "cn-northwest-1",
                    "dualstack": True,
                },
            ),
            ###
            Case(
                ("https://s3-fips.amazonaws.com", None),
                {
                    "s3_prefix": "s3-fips.",
                    "domain_suffix": "amazonaws.com",
                    "region": None,
                    "dualstack": False,
                },
            ),
            Case(
                ("https://s3-fips.amazonaws.com", "ap-south-1a"),
                {
                    "s3_prefix": "s3-fips.",
                    "domain_suffix": "amazonaws.com",
                    "region": "ap-south-1a",
                    "dualstack": False,
                },
            ),
            Case(
                ("https://s3-fips.us-gov-east-1.amazonaws.com", None),
                {
                    "s3_prefix": "s3-fips.",
                    "domain_suffix": "amazonaws.com",
                    "region": "us-gov-east-1",
                    "dualstack": False,
                },
            ),
            Case(
                ("https://s3-fips.me-south-1.amazonaws.com", "cn-northwest-1"),
                {
                    "s3_prefix": "s3-fips.",
                    "domain_suffix": "amazonaws.com",
                    "region": "cn-northwest-1",
                    "dualstack": False,
                },
            ),
            ###
            Case(
                ("https://s3-fips.dualstack.amazonaws.com", None),
                {
                    "s3_prefix": "s3-fips.",
                    "domain_suffix": "amazonaws.com",
                    "region": None,
                    "dualstack": True,
                },
            ),
            Case(
                ("https://s3-fips.dualstack.amazonaws.com", "ap-south-1a"),
                {
                    "s3_prefix": "s3-fips.",
                    "domain_suffix": "amazonaws.com",
                    "region": "ap-south-1a",
                    "dualstack": True,
                },
            ),
            Case(
                ("https://s3-fips.dualstack.us-gov-east-1.amazonaws.com", None),
                {
                    "s3_prefix": "s3-fips.",
                    "domain_suffix": "amazonaws.com",
                    "region": "us-gov-east-1",
                    "dualstack": True,
                },
            ),
            Case(
                ("https://s3-fips.dualstack.me-south-1.amazonaws.com",
                 "cn-northwest-1"),
                {
                    "s3_prefix": "s3-fips.",
                    "domain_suffix": "amazonaws.com",
                    "region": "cn-northwest-1",
                    "dualstack": True,
                },
            ),
            ###
            Case(
                ("https://s3-external-1.amazonaws.com", None),
                {
                    "s3_prefix": "s3-external-1.",
                    "domain_suffix": "amazonaws.com",
                    "region": "us-east-1",
                    "dualstack": False,
                },
            ),
            Case(
                ("https://s3-us-gov-west-1.amazonaws.com", None),
                {
                    "s3_prefix": "s3-us-gov-west-1.",
                    "domain_suffix": "amazonaws.com",
                    "region": "us-gov-west-1",
                    "dualstack": False,
                },
            ),
            Case(
                ("https://s3-fips-us-gov-west-1.amazonaws.com", None),
                {
                    "s3_prefix": "s3-fips-us-gov-west-1.",
                    "domain_suffix": "amazonaws.com",
                    "region": "us-gov-west-1",
                    "dualstack": False,
                },
            ),
            ###
            Case(
                ("https://bucket.vpce-1a2b3c4d-5e6f.s3.us-east-1."
                 "vpce.amazonaws.com", None),
                {
                    "s3_prefix": "bucket.vpce-1a2b3c4d-5e6f.s3.",
                    "domain_suffix": "vpce.amazonaws.com",
                    "region": "us-east-1",
                    "dualstack": False,
                },
            ),
            Case(
                ("https://accesspoint.vpce-1a2b3c4d-5e6f.s3.us-east-1."
                 "vpce.amazonaws.com", None),
                {
                    "s3_prefix": "accesspoint.vpce-1a2b3c4d-5e6f.s3.",
                    "domain_suffix": "vpce.amazonaws.com",
                    "region": "us-east-1",
                    "dualstack": False,
                },
            ),
            ###
            Case(
                ("https://012345678901.s3-control.amazonaws.com", None),
                {
                    "s3_prefix": "012345678901.s3-control.",
                    "domain_suffix": "amazonaws.com",
                    "region": None,
                    "dualstack": False,
                },
            ),
            Case(
                ("https://012345678901.s3-control.amazonaws.com",
                 "ap-south-1a"),
                {
                    "s3_prefix": "012345678901.s3-control.",
                    "domain_suffix": "amazonaws.com",
                    "region": "ap-south-1a",
                    "dualstack": False,
                },
            ),
            Case(
                ("https://012345678901.s3-control.us-gov-east-1.amazonaws.com",
                 None),
                {
                    "s3_prefix": "012345678901.s3-control.",
                    "domain_suffix": "amazonaws.com",
                    "region": "us-gov-east-1",
                    "dualstack": False,
                },
            ),
            Case(
                ("https://012345678901.s3-control.us-gov-east-1.amazonaws.com",
                 "cn-northwest-1"),
                {
                    "s3_prefix": "012345678901.s3-control.",
                    "domain_suffix": "amazonaws.com",
                    "region": "cn-northwest-1",
                    "dualstack": False,
                },
            ),
            ###
            Case(
                ("https://012345678901.s3-control.dualstack.amazonaws.com",
                 None),
                {
                    "s3_prefix": "012345678901.s3-control.",
                    "domain_suffix": "amazonaws.com",
                    "region": None,
                    "dualstack": True,
                },
            ),
            Case(
                ("https://012345678901.s3-control.dualstack.amazonaws.com",
                 "ap-south-1a"),
                {
                    "s3_prefix": "012345678901.s3-control.",
                    "domain_suffix": "amazonaws.com",
                    "region": "ap-south-1a",
                    "dualstack": True,
                },
            ),
            Case(
                ("https://012345678901.s3-control.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 None),
                {
                    "s3_prefix": "012345678901.s3-control.",
                    "domain_suffix": "amazonaws.com",
                    "region": "us-gov-east-1",
                    "dualstack": True,
                },
            ),
            Case(
                ("https://012345678901.s3-control.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 "cn-northwest-1"),
                {
                    "s3_prefix": "012345678901.s3-control.",
                    "domain_suffix": "amazonaws.com",
                    "region": "cn-northwest-1",
                    "dualstack": True,
                },
            ),
            ###
            Case(
                ("https://012345678901.s3-control-fips.amazonaws.com", None),
                {
                    "s3_prefix": "012345678901.s3-control-fips.",
                    "domain_suffix": "amazonaws.com",
                    "region": None,
                    "dualstack": False,
                },
            ),
            Case(
                ("https://012345678901.s3-control-fips.amazonaws.com",
                 "ap-south-1a"),
                {
                    "s3_prefix": "012345678901.s3-control-fips.",
                    "domain_suffix": "amazonaws.com",
                    "region": "ap-south-1a",
                    "dualstack": False,
                },
            ),
            Case(
                ("https://012345678901.s3-control-fips.us-gov-east-1."
                 "amazonaws.com",
                 None),
                {
                    "s3_prefix": "012345678901.s3-control-fips.",
                    "domain_suffix": "amazonaws.com",
                    "region": "us-gov-east-1",
                    "dualstack": False,
                },
            ),
            Case(
                ("https://012345678901.s3-control-fips.us-gov-east-1."
                 "amazonaws.com",
                 "cn-northwest-1"),
                {
                    "s3_prefix": "012345678901.s3-control-fips.",
                    "domain_suffix": "amazonaws.com",
                    "region": "cn-northwest-1",
                    "dualstack": False,
                },
            ),
            ###
            Case(
                ("https://012345678901.s3-control-fips.dualstack.amazonaws.com",
                 None),
                {
                    "s3_prefix": "012345678901.s3-control-fips.",
                    "domain_suffix": "amazonaws.com",
                    "region": None,
                    "dualstack": True,
                },
            ),
            Case(
                ("https://012345678901.s3-control-fips.dualstack.amazonaws.com",
                 "ap-south-1a"),
                {
                    "s3_prefix": "012345678901.s3-control-fips.",
                    "domain_suffix": "amazonaws.com",
                    "region": "ap-south-1a",
                    "dualstack": True,
                },
            ),
            Case(
                ("https://012345678901.s3-control-fips.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 None),
                {
                    "s3_prefix": "012345678901.s3-control-fips.",
                    "domain_suffix": "amazonaws.com",
                    "region": "us-gov-east-1",
                    "dualstack": True,
                },
            ),
            Case(
                ("https://012345678901.s3-control-fips.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 "cn-northwest-1"),
                {
                    "s3_prefix": "012345678901.s3-control-fips.",
                    "domain_suffix": "amazonaws.com",
                    "region": "cn-northwest-1",
                    "dualstack": True,
                },
            ),
            ###
            Case(
                ("https://s3-accesspoint.amazonaws.com", None),
                {
                    "s3_prefix": "s3-accesspoint.",
                    "domain_suffix": "amazonaws.com",
                    "region": None,
                    "dualstack": False,
                },
            ),
            Case(
                ("https://s3-accesspoint.amazonaws.com",
                 "ap-south-1a"),
                {
                    "s3_prefix": "s3-accesspoint.",
                    "domain_suffix": "amazonaws.com",
                    "region": "ap-south-1a",
                    "dualstack": False,
                },
            ),
            Case(
                ("https://s3-accesspoint.us-gov-east-1.amazonaws.com",
                 None),
                {
                    "s3_prefix": "s3-accesspoint.",
                    "domain_suffix": "amazonaws.com",
                    "region": "us-gov-east-1",
                    "dualstack": False,
                },
            ),
            Case(
                ("https://s3-accesspoint.us-gov-east-1.amazonaws.com",
                 "cn-northwest-1"),
                {
                    "s3_prefix": "s3-accesspoint.",
                    "domain_suffix": "amazonaws.com",
                    "region": "cn-northwest-1",
                    "dualstack": False,
                },
            ),
            ###
            Case(
                ("https://s3-accesspoint.dualstack.amazonaws.com",
                 None),
                {
                    "s3_prefix": "s3-accesspoint.",
                    "domain_suffix": "amazonaws.com",
                    "region": None,
                    "dualstack": True,
                },
            ),
            Case(
                ("https://s3-accesspoint.dualstack.amazonaws.com",
                 "ap-south-1a"),
                {
                    "s3_prefix": "s3-accesspoint.",
                    "domain_suffix": "amazonaws.com",
                    "region": "ap-south-1a",
                    "dualstack": True,
                },
            ),
            Case(
                ("https://s3-accesspoint.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 None),
                {
                    "s3_prefix": "s3-accesspoint.",
                    "domain_suffix": "amazonaws.com",
                    "region": "us-gov-east-1",
                    "dualstack": True,
                },
            ),
            Case(
                ("https://s3-accesspoint.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 "cn-northwest-1"),
                {
                    "s3_prefix": "s3-accesspoint.",
                    "domain_suffix": "amazonaws.com",
                    "region": "cn-northwest-1",
                    "dualstack": True,
                },
            ),
            ###
            Case(
                ("https://s3-accesspoint-fips.amazonaws.com", None),
                {
                    "s3_prefix": "s3-accesspoint-fips.",
                    "domain_suffix": "amazonaws.com",
                    "region": None,
                    "dualstack": False,
                },
            ),
            Case(
                ("https://s3-accesspoint-fips.amazonaws.com",
                 "ap-south-1a"),
                {
                    "s3_prefix": "s3-accesspoint-fips.",
                    "domain_suffix": "amazonaws.com",
                    "region": "ap-south-1a",
                    "dualstack": False,
                },
            ),
            Case(
                ("https://s3-accesspoint-fips.us-gov-east-1."
                 "amazonaws.com",
                 None),
                {
                    "s3_prefix": "s3-accesspoint-fips.",
                    "domain_suffix": "amazonaws.com",
                    "region": "us-gov-east-1",
                    "dualstack": False,
                },
            ),
            Case(
                ("https://s3-accesspoint-fips.us-gov-east-1."
                 "amazonaws.com",
                 "cn-northwest-1"),
                {
                    "s3_prefix": "s3-accesspoint-fips.",
                    "domain_suffix": "amazonaws.com",
                    "region": "cn-northwest-1",
                    "dualstack": False,
                },
            ),
            ###
            Case(
                ("https://s3-accesspoint-fips.dualstack.amazonaws.com",
                 None),
                {
                    "s3_prefix": "s3-accesspoint-fips.",
                    "domain_suffix": "amazonaws.com",
                    "region": None,
                    "dualstack": True,
                },
            ),
            Case(
                ("https://s3-accesspoint-fips.dualstack.amazonaws.com",
                 "ap-south-1a"),
                {
                    "s3_prefix": "s3-accesspoint-fips.",
                    "domain_suffix": "amazonaws.com",
                    "region": "ap-south-1a",
                    "dualstack": True,
                },
            ),
            Case(
                ("https://s3-accesspoint-fips.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 None),
                {
                    "s3_prefix": "s3-accesspoint-fips.",
                    "domain_suffix": "amazonaws.com",
                    "region": "us-gov-east-1",
                    "dualstack": True,
                },
            ),
            Case(
                ("https://s3-accesspoint-fips.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 "cn-northwest-1"),
                {
                    "s3_prefix": "s3-accesspoint-fips.",
                    "domain_suffix": "amazonaws.com",
                    "region": "cn-northwest-1",
                    "dualstack": True,
                },
            ),
            ###
            Case(
                ("https://my-load-balancer-1234567890.us-west-2.elb."
                 "amazonaws.com", "us-west-2"),
                None,
            ),
        ]

        for case in cases:
            url = BaseURL(*case.args)
            self.assertEqual(url._aws_info, case.result)

    def test_aws_list_buckets_build(self):
        Case = namedtuple("Case", ["args", "result"])
        cases = [
            Case(
                ("https://s3.amazonaws.com", None),
                "https://s3.us-east-1.amazonaws.com/",
            ),
            Case(
                ("https://s3.amazonaws.com", "ap-south-1a"),
                "https://s3.ap-south-1a.amazonaws.com/",
            ),
            Case(
                ("https://s3.us-gov-east-1.amazonaws.com", None),
                "https://s3.us-gov-east-1.amazonaws.com/",
            ),
            Case(
                ("https://s3.me-south-1.amazonaws.com", "cn-northwest-1"),
                "https://s3.cn-northwest-1.amazonaws.com/",
            ),
            ###
            Case(
                ("https://s3.dualstack.amazonaws.com", None),
                "https://s3.us-east-1.amazonaws.com/",
            ),
            Case(
                ("https://s3.dualstack.amazonaws.com", "ap-south-1a"),
                "https://s3.ap-south-1a.amazonaws.com/",
            ),
            Case(
                ("https://s3.dualstack.us-gov-east-1.amazonaws.com", None),
                "https://s3.us-gov-east-1.amazonaws.com/",
            ),
            Case(
                ("https://s3.dualstack.me-south-1.amazonaws.com",
                 "cn-northwest-1"),
                "https://s3.cn-northwest-1.amazonaws.com/",
            ),
            ###
            Case(
                ("https://s3-accelerate.amazonaws.com", None),
                "https://s3.us-east-1.amazonaws.com/",
            ),
            Case(
                ("https://s3-accelerate.amazonaws.com", "ap-south-1a"),
                "https://s3.ap-south-1a.amazonaws.com/",
            ),
            Case(
                ("https://s3-accelerate.us-gov-east-1.amazonaws.com", None),
                "https://s3.us-gov-east-1.amazonaws.com/",
            ),
            Case(
                ("https://s3-accelerate.me-south-1.amazonaws.com",
                 "cn-northwest-1"),
                "https://s3.cn-northwest-1.amazonaws.com/",
            ),
            ###
            Case(
                ("https://s3-accelerate.dualstack.amazonaws.com", None),
                "https://s3.us-east-1.amazonaws.com/",
            ),
            Case(
                ("https://s3-accelerate.dualstack.amazonaws.com",
                 "ap-south-1a"),
                "https://s3.ap-south-1a.amazonaws.com/",
            ),
            Case(
                ("https://s3-accelerate.dualstack.us-gov-east-1.amazonaws.com",
                 None),
                "https://s3.us-gov-east-1.amazonaws.com/",
            ),
            Case(
                ("https://s3-accelerate.dualstack.me-south-1.amazonaws.com",
                 "cn-northwest-1"),
                "https://s3.cn-northwest-1.amazonaws.com/",
            ),
            ###
            Case(
                ("https://s3-fips.amazonaws.com", None),
                "https://s3.us-east-1.amazonaws.com/",
            ),
            Case(
                ("https://s3-fips.amazonaws.com", "ap-south-1a"),
                "https://s3.ap-south-1a.amazonaws.com/",
            ),
            Case(
                ("https://s3-fips.us-gov-east-1.amazonaws.com", None),
                "https://s3.us-gov-east-1.amazonaws.com/",
            ),
            Case(
                ("https://s3-fips.me-south-1.amazonaws.com", "cn-northwest-1"),
                "https://s3.cn-northwest-1.amazonaws.com/",
            ),
            ###
            Case(
                ("https://s3-fips.dualstack.amazonaws.com", None),
                "https://s3.us-east-1.amazonaws.com/",
            ),
            Case(
                ("https://s3-fips.dualstack.amazonaws.com", "ap-south-1a"),
                "https://s3.ap-south-1a.amazonaws.com/",
            ),
            Case(
                ("https://s3-fips.dualstack.us-gov-east-1.amazonaws.com", None),
                "https://s3.us-gov-east-1.amazonaws.com/",
            ),
            Case(
                ("https://s3-fips.dualstack.me-south-1.amazonaws.com",
                 "cn-northwest-1"),
                "https://s3.cn-northwest-1.amazonaws.com/",
            ),
            ###
            Case(
                ("https://s3-external-1.amazonaws.com", None),
                "https://s3-external-1.amazonaws.com/",
            ),
            Case(
                ("https://s3-us-gov-west-1.amazonaws.com", None),
                "https://s3-us-gov-west-1.amazonaws.com/",
            ),
            Case(
                ("https://s3-fips-us-gov-west-1.amazonaws.com", None),
                "https://s3-fips-us-gov-west-1.amazonaws.com/",
            ),
            ###
            Case(
                ("https://bucket.vpce-1a2b3c4d-5e6f.s3.us-east-1."
                 "vpce.amazonaws.com", None),
                "https://bucket.vpce-1a2b3c4d-5e6f.s3.us-east-1."
                "vpce.amazonaws.com/",
            ),
            Case(
                ("https://accesspoint.vpce-1a2b3c4d-5e6f.s3.us-east-1."
                 "vpce.amazonaws.com", None),
                "https://accesspoint.vpce-1a2b3c4d-5e6f.s3.us-east-1."
                "vpce.amazonaws.com/",
            ),
            ###
            Case(
                ("https://012345678901.s3-control.amazonaws.com", None),
                "https://012345678901.s3-control.us-east-1.amazonaws.com/",
            ),
            Case(
                ("https://012345678901.s3-control.amazonaws.com",
                 "ap-south-1a"),
                "https://012345678901.s3-control.ap-south-1a.amazonaws.com/",
            ),
            Case(
                ("https://012345678901.s3-control.us-gov-east-1.amazonaws.com",
                 None),
                "https://012345678901.s3-control.us-gov-east-1.amazonaws.com/",
            ),
            Case(
                ("https://012345678901.s3-control.us-gov-east-1.amazonaws.com",
                 "cn-northwest-1"),
                "https://012345678901.s3-control.cn-northwest-1.amazonaws.com/",
            ),
            ###
            Case(
                ("https://012345678901.s3-control.dualstack.amazonaws.com",
                 None),
                "https://012345678901.s3-control.us-east-1."
                "amazonaws.com/",
            ),
            Case(
                ("https://012345678901.s3-control.dualstack.amazonaws.com",
                 "ap-south-1a"),
                "https://012345678901.s3-control.ap-south-1a."
                "amazonaws.com/",
            ),
            Case(
                ("https://012345678901.s3-control.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 None),
                "https://012345678901.s3-control.us-gov-east-1."
                "amazonaws.com/",
            ),
            Case(
                ("https://012345678901.s3-control.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 "cn-northwest-1"),
                "https://012345678901.s3-control.cn-northwest-1."
                "amazonaws.com/",
            ),
            ###
            Case(
                ("https://012345678901.s3-control-fips.amazonaws.com", None),
                "https://012345678901.s3-control-fips.us-east-1.amazonaws.com/",
            ),
            Case(
                ("https://012345678901.s3-control-fips.amazonaws.com",
                 "ap-south-1a"),
                "https://012345678901.s3-control-fips.ap-south-1a."
                "amazonaws.com/",
            ),
            Case(
                ("https://012345678901.s3-control-fips.us-gov-east-1."
                 "amazonaws.com",
                 None),
                "https://012345678901.s3-control-fips.us-gov-east-1."
                "amazonaws.com/",
            ),
            Case(
                ("https://012345678901.s3-control-fips.us-gov-east-1."
                 "amazonaws.com",
                 "cn-northwest-1"),
                "https://012345678901.s3-control-fips.cn-northwest-1."
                "amazonaws.com/",
            ),
            ###
            Case(
                ("https://012345678901.s3-control-fips.dualstack.amazonaws.com",
                 None),
                "https://012345678901.s3-control-fips.us-east-1."
                "amazonaws.com/",
            ),
            Case(
                ("https://012345678901.s3-control-fips.dualstack.amazonaws.com",
                 "ap-south-1a"),
                "https://012345678901.s3-control-fips.ap-south-1a."
                "amazonaws.com/",
            ),
            Case(
                ("https://012345678901.s3-control-fips.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 None),
                "https://012345678901.s3-control-fips.us-gov-east-1."
                "amazonaws.com/",
            ),
            Case(
                ("https://012345678901.s3-control-fips.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 "cn-northwest-1"),
                "https://012345678901.s3-control-fips.cn-northwest-1."
                "amazonaws.com/",
            ),
            ###
            Case(
                ("https://s3-accesspoint.amazonaws.com", None),
                "https://s3.us-east-1.amazonaws.com/",
            ),
            Case(
                ("https://s3-accesspoint.amazonaws.com",
                 "ap-south-1a"),
                "https://s3.ap-south-1a.amazonaws.com/",
            ),
            Case(
                ("https://s3-accesspoint.us-gov-east-1.amazonaws.com",
                 None),
                "https://s3.us-gov-east-1.amazonaws.com/",
            ),
            Case(
                ("https://s3-accesspoint.us-gov-east-1.amazonaws.com",
                 "cn-northwest-1"),
                "https://s3.cn-northwest-1.amazonaws.com/",
            ),
            ###
            Case(
                ("https://s3-accesspoint.dualstack.amazonaws.com",
                 None),
                "https://s3.us-east-1.amazonaws.com/",
            ),
            Case(
                ("https://s3-accesspoint.dualstack.amazonaws.com",
                 "ap-south-1a"),
                "https://s3.ap-south-1a.amazonaws.com/",
            ),
            Case(
                ("https://s3-accesspoint.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 None),
                "https://s3.us-gov-east-1.amazonaws.com/",
            ),
            Case(
                ("https://s3-accesspoint.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 "cn-northwest-1"),
                "https://s3.cn-northwest-1.amazonaws.com/",
            ),
            ###
            Case(
                ("https://s3-accesspoint-fips.amazonaws.com", None),
                "https://s3.us-east-1.amazonaws.com/",
            ),
            Case(
                ("https://s3-accesspoint-fips.amazonaws.com",
                 "ap-south-1a"),
                "https://s3.ap-south-1a.amazonaws.com/",
            ),
            Case(
                ("https://s3-accesspoint-fips.us-gov-east-1."
                 "amazonaws.com",
                 None),
                "https://s3.us-gov-east-1.amazonaws.com/",
            ),
            Case(
                ("https://s3-accesspoint-fips.us-gov-east-1."
                 "amazonaws.com",
                 "cn-northwest-1"),
                "https://s3.cn-northwest-1.amazonaws.com/",
            ),
            ###
            Case(
                ("https://s3-accesspoint-fips.dualstack.amazonaws.com",
                 None),
                "https://s3.us-east-1.amazonaws.com/",
            ),
            Case(
                ("https://s3-accesspoint-fips.dualstack.amazonaws.com",
                 "ap-south-1a"),
                "https://s3.ap-south-1a.amazonaws.com/",
            ),
            Case(
                ("https://s3-accesspoint-fips.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 None),
                "https://s3.us-gov-east-1.amazonaws.com/",
            ),
            Case(
                ("https://s3-accesspoint-fips.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 "cn-northwest-1"),
                "https://s3.cn-northwest-1.amazonaws.com/",
            ),
            ###
            Case(
                ("https://my-load-balancer-1234567890.us-west-2.elb."
                 "amazonaws.com", "us-west-2"),
                "https://my-load-balancer-1234567890.us-west-2.elb."
                "amazonaws.com/",
            ),
        ]

        for case in cases:
            base_url = BaseURL(*case.args)
            url = urlunsplit(base_url.build(
                "GET", base_url.region or "us-east-1"))
            self.assertEqual(str(url), case.result)

    def test_aws_bucket_build(self):
        Case = namedtuple("Case", ["args", "result"])
        cases = [
            Case(
                ("https://s3.amazonaws.com", None),
                "https://my-bucket.s3.us-east-1.amazonaws.com/",
            ),
            Case(
                ("https://s3.amazonaws.com", "ap-south-1a"),
                "https://my-bucket.s3.ap-south-1a.amazonaws.com/",
            ),
            Case(
                ("https://s3.us-gov-east-1.amazonaws.com", None),
                "https://my-bucket.s3.us-gov-east-1.amazonaws.com/",
            ),
            Case(
                ("https://s3.me-south-1.amazonaws.com", "cn-northwest-1"),
                "https://my-bucket.s3.cn-northwest-1.amazonaws.com/",
            ),
            ###
            Case(
                ("https://s3.dualstack.amazonaws.com", None),
                "https://my-bucket.s3.dualstack.us-east-1.amazonaws.com/",
            ),
            Case(
                ("https://s3.dualstack.amazonaws.com", "ap-south-1a"),
                "https://my-bucket.s3.dualstack.ap-south-1a.amazonaws.com/",
            ),
            Case(
                ("https://s3.dualstack.us-gov-east-1.amazonaws.com", None),
                "https://my-bucket.s3.dualstack.us-gov-east-1.amazonaws.com/",
            ),
            Case(
                ("https://s3.dualstack.me-south-1.amazonaws.com",
                 "cn-northwest-1"),
                "https://my-bucket.s3.dualstack.cn-northwest-1.amazonaws.com/",
            ),
            ###
            Case(
                ("https://s3-accelerate.amazonaws.com", None),
                "https://my-bucket.s3-accelerate.amazonaws.com/",
            ),
            Case(
                ("https://s3-accelerate.amazonaws.com", "ap-south-1a"),
                "https://my-bucket.s3-accelerate.amazonaws.com/",
            ),
            Case(
                ("https://s3-accelerate.us-gov-east-1.amazonaws.com", None),
                "https://my-bucket.s3-accelerate.amazonaws.com/",
            ),
            Case(
                ("https://s3-accelerate.me-south-1.amazonaws.com",
                 "cn-northwest-1"),
                "https://my-bucket.s3-accelerate.amazonaws.com/",
            ),
            ###
            Case(
                ("https://s3-accelerate.dualstack.amazonaws.com", None),
                "https://my-bucket.s3-accelerate.dualstack.amazonaws.com/",
            ),
            Case(
                ("https://s3-accelerate.dualstack.amazonaws.com",
                 "ap-south-1a"),
                "https://my-bucket.s3-accelerate.dualstack.amazonaws.com/",
            ),
            Case(
                ("https://s3-accelerate.dualstack.us-gov-east-1.amazonaws.com",
                 None),
                "https://my-bucket.s3-accelerate.dualstack.amazonaws.com/",
            ),
            Case(
                ("https://s3-accelerate.dualstack.me-south-1.amazonaws.com",
                 "cn-northwest-1"),
                "https://my-bucket.s3-accelerate.dualstack.amazonaws.com/",
            ),
            ###
            Case(
                ("https://s3-fips.amazonaws.com", None),
                "https://my-bucket.s3-fips.us-east-1.amazonaws.com/",
            ),
            Case(
                ("https://s3-fips.amazonaws.com", "ap-south-1a"),
                "https://my-bucket.s3-fips.ap-south-1a.amazonaws.com/",
            ),
            Case(
                ("https://s3-fips.us-gov-east-1.amazonaws.com", None),
                "https://my-bucket.s3-fips.us-gov-east-1.amazonaws.com/",
            ),
            Case(
                ("https://s3-fips.me-south-1.amazonaws.com", "cn-northwest-1"),
                "https://my-bucket.s3-fips.cn-northwest-1.amazonaws.com/",
            ),
            ###
            Case(
                ("https://s3-fips.dualstack.amazonaws.com", None),
                "https://my-bucket.s3-fips.dualstack.us-east-1.amazonaws.com/",
            ),
            Case(
                ("https://s3-fips.dualstack.amazonaws.com", "ap-south-1a"),
                "https://my-bucket.s3-fips.dualstack.ap-south-1a."
                "amazonaws.com/",
            ),
            Case(
                ("https://s3-fips.dualstack.us-gov-east-1.amazonaws.com", None),
                "https://my-bucket.s3-fips.dualstack.us-gov-east-1."
                "amazonaws.com/",
            ),
            Case(
                ("https://s3-fips.dualstack.me-south-1.amazonaws.com",
                 "cn-northwest-1"),
                "https://my-bucket.s3-fips.dualstack.cn-northwest-1."
                "amazonaws.com/",
            ),
            ###
            Case(
                ("https://s3-external-1.amazonaws.com", None),
                "https://my-bucket.s3-external-1.amazonaws.com/",
            ),
            Case(
                ("https://s3-us-gov-west-1.amazonaws.com", None),
                "https://my-bucket.s3-us-gov-west-1.amazonaws.com/",
            ),
            Case(
                ("https://s3-fips-us-gov-west-1.amazonaws.com", None),
                "https://my-bucket.s3-fips-us-gov-west-1.amazonaws.com/",
            ),
            ###
            Case(
                ("https://bucket.vpce-1a2b3c4d-5e6f.s3.us-east-1."
                 "vpce.amazonaws.com", None),
                "https://my-bucket.bucket.vpce-1a2b3c4d-5e6f.s3.us-east-1."
                "vpce.amazonaws.com/",
            ),
            Case(
                ("https://accesspoint.vpce-1a2b3c4d-5e6f.s3.us-east-1."
                 "vpce.amazonaws.com", None),
                "https://my-bucket.accesspoint.vpce-1a2b3c4d-5e6f.s3.us-east-1."
                "vpce.amazonaws.com/",
            ),
            ###
            Case(
                ("https://012345678901.s3-control.amazonaws.com", None),
                "https://my-bucket.012345678901.s3-control.us-east-1."
                "amazonaws.com/",
            ),
            Case(
                ("https://012345678901.s3-control.amazonaws.com",
                 "ap-south-1a"),
                "https://my-bucket.012345678901.s3-control.ap-south-1a."
                "amazonaws.com/",
            ),
            Case(
                ("https://012345678901.s3-control.us-gov-east-1.amazonaws.com",
                 None),
                "https://my-bucket.012345678901.s3-control.us-gov-east-1."
                "amazonaws.com/",
            ),
            Case(
                ("https://012345678901.s3-control.us-gov-east-1.amazonaws.com",
                 "cn-northwest-1"),
                "https://my-bucket.012345678901.s3-control.cn-northwest-1."
                "amazonaws.com/",
            ),
            ###
            Case(
                ("https://012345678901.s3-control.dualstack.amazonaws.com",
                 None),
                "https://my-bucket.012345678901.s3-control.dualstack.us-east-1."
                "amazonaws.com/",
            ),
            Case(
                ("https://012345678901.s3-control.dualstack.amazonaws.com",
                 "ap-south-1a"),
                "https://my-bucket.012345678901.s3-control.dualstack."
                "ap-south-1a.amazonaws.com/",
            ),
            Case(
                ("https://012345678901.s3-control.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 None),
                "https://my-bucket.012345678901.s3-control.dualstack."
                "us-gov-east-1.amazonaws.com/",
            ),
            Case(
                ("https://012345678901.s3-control.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 "cn-northwest-1"),
                "https://my-bucket.012345678901.s3-control.dualstack."
                "cn-northwest-1.amazonaws.com/",
            ),
            ###
            Case(
                ("https://012345678901.s3-control-fips.amazonaws.com", None),
                "https://my-bucket.012345678901.s3-control-fips.us-east-1."
                "amazonaws.com/",
            ),
            Case(
                ("https://012345678901.s3-control-fips.amazonaws.com",
                 "ap-south-1a"),
                "https://my-bucket.012345678901.s3-control-fips.ap-south-1a."
                "amazonaws.com/",
            ),
            Case(
                ("https://012345678901.s3-control-fips.us-gov-east-1."
                 "amazonaws.com",
                 None),
                "https://my-bucket.012345678901.s3-control-fips.us-gov-east-1."
                "amazonaws.com/",
            ),
            Case(
                ("https://012345678901.s3-control-fips.us-gov-east-1."
                 "amazonaws.com",
                 "cn-northwest-1"),
                "https://my-bucket.012345678901.s3-control-fips.cn-northwest-1."
                "amazonaws.com/",
            ),
            ###
            Case(
                ("https://012345678901.s3-control-fips.dualstack.amazonaws.com",
                 None),
                "https://my-bucket.012345678901.s3-control-fips.dualstack."
                "us-east-1.amazonaws.com/",
            ),
            Case(
                ("https://012345678901.s3-control-fips.dualstack.amazonaws.com",
                 "ap-south-1a"),
                "https://my-bucket.012345678901.s3-control-fips.dualstack."
                "ap-south-1a.amazonaws.com/",
            ),
            Case(
                ("https://012345678901.s3-control-fips.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 None),
                "https://my-bucket.012345678901.s3-control-fips.dualstack."
                "us-gov-east-1.amazonaws.com/",
            ),
            Case(
                ("https://012345678901.s3-control-fips.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 "cn-northwest-1"),
                "https://my-bucket.012345678901.s3-control-fips.dualstack."
                "cn-northwest-1.amazonaws.com/",
            ),
            ###
            Case(
                ("https://s3-accesspoint.amazonaws.com", None),
                "https://my-bucket.s3-accesspoint.us-east-1.amazonaws.com/",
            ),
            Case(
                ("https://s3-accesspoint.amazonaws.com",
                 "ap-south-1a"),
                "https://my-bucket.s3-accesspoint.ap-south-1a.amazonaws.com/",
            ),
            Case(
                ("https://s3-accesspoint.us-gov-east-1.amazonaws.com",
                 None),
                "https://my-bucket.s3-accesspoint.us-gov-east-1.amazonaws.com/",
            ),
            Case(
                ("https://s3-accesspoint.us-gov-east-1.amazonaws.com",
                 "cn-northwest-1"),
                "https://my-bucket.s3-accesspoint.cn-northwest-1."
                "amazonaws.com/",
            ),
            ###
            Case(
                ("https://s3-accesspoint.dualstack.amazonaws.com",
                 None),
                "https://my-bucket.s3-accesspoint.dualstack.us-east-1."
                "amazonaws.com/",
            ),
            Case(
                ("https://s3-accesspoint.dualstack.amazonaws.com",
                 "ap-south-1a"),
                "https://my-bucket.s3-accesspoint.dualstack.ap-south-1a."
                "amazonaws.com/",
            ),
            Case(
                ("https://s3-accesspoint.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 None),
                "https://my-bucket.s3-accesspoint.dualstack.us-gov-east-1."
                "amazonaws.com/",
            ),
            Case(
                ("https://s3-accesspoint.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 "cn-northwest-1"),
                "https://my-bucket.s3-accesspoint.dualstack.cn-northwest-1."
                "amazonaws.com/",
            ),
            ###
            Case(
                ("https://s3-accesspoint-fips.amazonaws.com", None),
                "https://my-bucket.s3-accesspoint-fips.us-east-1."
                "amazonaws.com/",
            ),
            Case(
                ("https://s3-accesspoint-fips.amazonaws.com",
                 "ap-south-1a"),
                "https://my-bucket.s3-accesspoint-fips.ap-south-1a."
                "amazonaws.com/",
            ),
            Case(
                ("https://s3-accesspoint-fips.us-gov-east-1."
                 "amazonaws.com",
                 None),
                "https://my-bucket.s3-accesspoint-fips.us-gov-east-1."
                "amazonaws.com/",
            ),
            Case(
                ("https://s3-accesspoint-fips.us-gov-east-1."
                 "amazonaws.com",
                 "cn-northwest-1"),
                "https://my-bucket.s3-accesspoint-fips.cn-northwest-1."
                "amazonaws.com/",
            ),
            ###
            Case(
                ("https://s3-accesspoint-fips.dualstack.amazonaws.com",
                 None),
                "https://my-bucket.s3-accesspoint-fips.dualstack.us-east-1."
                "amazonaws.com/",
            ),
            Case(
                ("https://s3-accesspoint-fips.dualstack.amazonaws.com",
                 "ap-south-1a"),
                "https://my-bucket.s3-accesspoint-fips.dualstack.ap-south-1a."
                "amazonaws.com/",
            ),
            Case(
                ("https://s3-accesspoint-fips.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 None),
                "https://my-bucket.s3-accesspoint-fips.dualstack.us-gov-east-1."
                "amazonaws.com/",
            ),
            Case(
                ("https://s3-accesspoint-fips.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 "cn-northwest-1"),
                "https://my-bucket.s3-accesspoint-fips.dualstack."
                "cn-northwest-1.amazonaws.com/",
            ),
            ###
            Case(
                ("https://my-load-balancer-1234567890.us-west-2.elb."
                 "amazonaws.com", "us-west-2"),
                "https://my-load-balancer-1234567890.us-west-2.elb."
                "amazonaws.com/my-bucket",
            ),
        ]

        for case in cases:
            base_url = BaseURL(*case.args)
            url = urlunsplit(base_url.build(
                "GET", base_url.region or "us-east-1", bucket_name="my-bucket"))
            self.assertEqual(str(url), case.result)

    def test_aws_object_build(self):
        Case = namedtuple("Case", ["args", "result"])
        cases = [
            Case(
                ("https://s3.amazonaws.com", None),
                "https://my-bucket.s3.us-east-1.amazonaws.com/"
                "path/to/my/object",
            ),
            Case(
                ("https://s3.amazonaws.com", "ap-south-1a"),
                "https://my-bucket.s3.ap-south-1a.amazonaws.com/"
                "path/to/my/object",
            ),
            Case(
                ("https://s3.us-gov-east-1.amazonaws.com", None),
                "https://my-bucket.s3.us-gov-east-1.amazonaws.com/"
                "path/to/my/object",
            ),
            Case(
                ("https://s3.me-south-1.amazonaws.com", "cn-northwest-1"),
                "https://my-bucket.s3.cn-northwest-1.amazonaws.com/"
                "path/to/my/object",
            ),
            ###
            Case(
                ("https://s3.dualstack.amazonaws.com", None),
                "https://my-bucket.s3.dualstack.us-east-1.amazonaws.com/"
                "path/to/my/object",
            ),
            Case(
                ("https://s3.dualstack.amazonaws.com", "ap-south-1a"),
                "https://my-bucket.s3.dualstack.ap-south-1a.amazonaws.com/"
                "path/to/my/object",
            ),
            Case(
                ("https://s3.dualstack.us-gov-east-1.amazonaws.com", None),
                "https://my-bucket.s3.dualstack.us-gov-east-1.amazonaws.com/"
                "path/to/my/object",
            ),
            Case(
                ("https://s3.dualstack.me-south-1.amazonaws.com",
                 "cn-northwest-1"),
                "https://my-bucket.s3.dualstack.cn-northwest-1.amazonaws.com/"
                "path/to/my/object",
            ),
            ###
            Case(
                ("https://s3-accelerate.amazonaws.com", None),
                "https://my-bucket.s3-accelerate.amazonaws.com/"
                "path/to/my/object",
            ),
            Case(
                ("https://s3-accelerate.amazonaws.com", "ap-south-1a"),
                "https://my-bucket.s3-accelerate.amazonaws.com/"
                "path/to/my/object",
            ),
            Case(
                ("https://s3-accelerate.us-gov-east-1.amazonaws.com", None),
                "https://my-bucket.s3-accelerate.amazonaws.com/"
                "path/to/my/object",
            ),
            Case(
                ("https://s3-accelerate.me-south-1.amazonaws.com",
                 "cn-northwest-1"),
                "https://my-bucket.s3-accelerate.amazonaws.com/"
                "path/to/my/object",
            ),
            ###
            Case(
                ("https://s3-accelerate.dualstack.amazonaws.com", None),
                "https://my-bucket.s3-accelerate.dualstack.amazonaws.com/"
                "path/to/my/object",
            ),
            Case(
                ("https://s3-accelerate.dualstack.amazonaws.com",
                 "ap-south-1a"),
                "https://my-bucket.s3-accelerate.dualstack.amazonaws.com/"
                "path/to/my/object",
            ),
            Case(
                ("https://s3-accelerate.dualstack.us-gov-east-1.amazonaws.com",
                 None),
                "https://my-bucket.s3-accelerate.dualstack.amazonaws.com/"
                "path/to/my/object",
            ),
            Case(
                ("https://s3-accelerate.dualstack.me-south-1.amazonaws.com",
                 "cn-northwest-1"),
                "https://my-bucket.s3-accelerate.dualstack.amazonaws.com/"
                "path/to/my/object",
            ),
            ###
            Case(
                ("https://s3-fips.amazonaws.com", None),
                "https://my-bucket.s3-fips.us-east-1.amazonaws.com/"
                "path/to/my/object",
            ),
            Case(
                ("https://s3-fips.amazonaws.com", "ap-south-1a"),
                "https://my-bucket.s3-fips.ap-south-1a.amazonaws.com/"
                "path/to/my/object",
            ),
            Case(
                ("https://s3-fips.us-gov-east-1.amazonaws.com", None),
                "https://my-bucket.s3-fips.us-gov-east-1.amazonaws.com/"
                "path/to/my/object",
            ),
            Case(
                ("https://s3-fips.me-south-1.amazonaws.com", "cn-northwest-1"),
                "https://my-bucket.s3-fips.cn-northwest-1.amazonaws.com/"
                "path/to/my/object",
            ),
            ###
            Case(
                ("https://s3-fips.dualstack.amazonaws.com", None),
                "https://my-bucket.s3-fips.dualstack.us-east-1.amazonaws.com/"
                "path/to/my/object",
            ),
            Case(
                ("https://s3-fips.dualstack.amazonaws.com", "ap-south-1a"),
                "https://my-bucket.s3-fips.dualstack.ap-south-1a."
                "amazonaws.com/path/to/my/object",
            ),
            Case(
                ("https://s3-fips.dualstack.us-gov-east-1.amazonaws.com", None),
                "https://my-bucket.s3-fips.dualstack.us-gov-east-1."
                "amazonaws.com/path/to/my/object",
            ),
            Case(
                ("https://s3-fips.dualstack.me-south-1.amazonaws.com",
                 "cn-northwest-1"),
                "https://my-bucket.s3-fips.dualstack.cn-northwest-1."
                "amazonaws.com/path/to/my/object",
            ),
            ###
            Case(
                ("https://s3-external-1.amazonaws.com", None),
                "https://my-bucket.s3-external-1.amazonaws.com/"
                "path/to/my/object",
            ),
            Case(
                ("https://s3-us-gov-west-1.amazonaws.com", None),
                "https://my-bucket.s3-us-gov-west-1.amazonaws.com/"
                "path/to/my/object",
            ),
            Case(
                ("https://s3-fips-us-gov-west-1.amazonaws.com", None),
                "https://my-bucket.s3-fips-us-gov-west-1.amazonaws.com/"
                "path/to/my/object",
            ),
            ###
            Case(
                ("https://bucket.vpce-1a2b3c4d-5e6f.s3.us-east-1."
                 "vpce.amazonaws.com", None),
                "https://my-bucket.bucket.vpce-1a2b3c4d-5e6f.s3.us-east-1."
                "vpce.amazonaws.com/path/to/my/object",
            ),
            Case(
                ("https://accesspoint.vpce-1a2b3c4d-5e6f.s3.us-east-1."
                 "vpce.amazonaws.com", None),
                "https://my-bucket.accesspoint.vpce-1a2b3c4d-5e6f.s3.us-east-1."
                "vpce.amazonaws.com/path/to/my/object",
            ),
            ###
            Case(
                ("https://012345678901.s3-control.amazonaws.com", None),
                "https://my-bucket.012345678901.s3-control.us-east-1."
                "amazonaws.com/path/to/my/object",
            ),
            Case(
                ("https://012345678901.s3-control.amazonaws.com",
                 "ap-south-1a"),
                "https://my-bucket.012345678901.s3-control.ap-south-1a."
                "amazonaws.com/path/to/my/object",
            ),
            Case(
                ("https://012345678901.s3-control.us-gov-east-1.amazonaws.com",
                 None),
                "https://my-bucket.012345678901.s3-control.us-gov-east-1."
                "amazonaws.com/path/to/my/object",
            ),
            Case(
                ("https://012345678901.s3-control.us-gov-east-1.amazonaws.com",
                 "cn-northwest-1"),
                "https://my-bucket.012345678901.s3-control.cn-northwest-1."
                "amazonaws.com/path/to/my/object",
            ),
            ###
            Case(
                ("https://012345678901.s3-control.dualstack.amazonaws.com",
                 None),
                "https://my-bucket.012345678901.s3-control.dualstack.us-east-1."
                "amazonaws.com/path/to/my/object",
            ),
            Case(
                ("https://012345678901.s3-control.dualstack.amazonaws.com",
                 "ap-south-1a"),
                "https://my-bucket.012345678901.s3-control.dualstack."
                "ap-south-1a.amazonaws.com/path/to/my/object",
            ),
            Case(
                ("https://012345678901.s3-control.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 None),
                "https://my-bucket.012345678901.s3-control.dualstack."
                "us-gov-east-1.amazonaws.com/path/to/my/object",
            ),
            Case(
                ("https://012345678901.s3-control.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 "cn-northwest-1"),
                "https://my-bucket.012345678901.s3-control.dualstack."
                "cn-northwest-1.amazonaws.com/path/to/my/object",
            ),
            ###
            Case(
                ("https://012345678901.s3-control-fips.amazonaws.com", None),
                "https://my-bucket.012345678901.s3-control-fips.us-east-1."
                "amazonaws.com/path/to/my/object",
            ),
            Case(
                ("https://012345678901.s3-control-fips.amazonaws.com",
                 "ap-south-1a"),
                "https://my-bucket.012345678901.s3-control-fips.ap-south-1a."
                "amazonaws.com/path/to/my/object",
            ),
            Case(
                ("https://012345678901.s3-control-fips.us-gov-east-1."
                 "amazonaws.com",
                 None),
                "https://my-bucket.012345678901.s3-control-fips.us-gov-east-1."
                "amazonaws.com/path/to/my/object",
            ),
            Case(
                ("https://012345678901.s3-control-fips.us-gov-east-1."
                 "amazonaws.com",
                 "cn-northwest-1"),
                "https://my-bucket.012345678901.s3-control-fips.cn-northwest-1."
                "amazonaws.com/path/to/my/object",
            ),
            ###
            Case(
                ("https://012345678901.s3-control-fips.dualstack.amazonaws.com",
                 None),
                "https://my-bucket.012345678901.s3-control-fips.dualstack."
                "us-east-1.amazonaws.com/path/to/my/object",
            ),
            Case(
                ("https://012345678901.s3-control-fips.dualstack.amazonaws.com",
                 "ap-south-1a"),
                "https://my-bucket.012345678901.s3-control-fips.dualstack."
                "ap-south-1a.amazonaws.com/path/to/my/object",
            ),
            Case(
                ("https://012345678901.s3-control-fips.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 None),
                "https://my-bucket.012345678901.s3-control-fips.dualstack."
                "us-gov-east-1.amazonaws.com/path/to/my/object",
            ),
            Case(
                ("https://012345678901.s3-control-fips.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 "cn-northwest-1"),
                "https://my-bucket.012345678901.s3-control-fips.dualstack."
                "cn-northwest-1.amazonaws.com/path/to/my/object",
            ),
            ###
            Case(
                ("https://s3-accesspoint.amazonaws.com", None),
                "https://my-bucket.s3-accesspoint.us-east-1.amazonaws.com/"
                "path/to/my/object",
            ),
            Case(
                ("https://s3-accesspoint.amazonaws.com",
                 "ap-south-1a"),
                "https://my-bucket.s3-accesspoint.ap-south-1a.amazonaws.com/"
                "path/to/my/object",
            ),
            Case(
                ("https://s3-accesspoint.us-gov-east-1.amazonaws.com",
                 None),
                "https://my-bucket.s3-accesspoint.us-gov-east-1.amazonaws.com/"
                "path/to/my/object",
            ),
            Case(
                ("https://s3-accesspoint.us-gov-east-1.amazonaws.com",
                 "cn-northwest-1"),
                "https://my-bucket.s3-accesspoint.cn-northwest-1."
                "amazonaws.com/path/to/my/object",
            ),
            ###
            Case(
                ("https://s3-accesspoint.dualstack.amazonaws.com",
                 None),
                "https://my-bucket.s3-accesspoint.dualstack.us-east-1."
                "amazonaws.com/path/to/my/object",
            ),
            Case(
                ("https://s3-accesspoint.dualstack.amazonaws.com",
                 "ap-south-1a"),
                "https://my-bucket.s3-accesspoint.dualstack.ap-south-1a."
                "amazonaws.com/path/to/my/object",
            ),
            Case(
                ("https://s3-accesspoint.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 None),
                "https://my-bucket.s3-accesspoint.dualstack.us-gov-east-1."
                "amazonaws.com/path/to/my/object",
            ),
            Case(
                ("https://s3-accesspoint.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 "cn-northwest-1"),
                "https://my-bucket.s3-accesspoint.dualstack.cn-northwest-1."
                "amazonaws.com/path/to/my/object",
            ),
            ###
            Case(
                ("https://s3-accesspoint-fips.amazonaws.com", None),
                "https://my-bucket.s3-accesspoint-fips.us-east-1."
                "amazonaws.com/path/to/my/object",
            ),
            Case(
                ("https://s3-accesspoint-fips.amazonaws.com",
                 "ap-south-1a"),
                "https://my-bucket.s3-accesspoint-fips.ap-south-1a."
                "amazonaws.com/path/to/my/object",
            ),
            Case(
                ("https://s3-accesspoint-fips.us-gov-east-1."
                 "amazonaws.com",
                 None),
                "https://my-bucket.s3-accesspoint-fips.us-gov-east-1."
                "amazonaws.com/path/to/my/object",
            ),
            Case(
                ("https://s3-accesspoint-fips.us-gov-east-1."
                 "amazonaws.com",
                 "cn-northwest-1"),
                "https://my-bucket.s3-accesspoint-fips.cn-northwest-1."
                "amazonaws.com/path/to/my/object",
            ),
            ###
            Case(
                ("https://s3-accesspoint-fips.dualstack.amazonaws.com",
                 None),
                "https://my-bucket.s3-accesspoint-fips.dualstack.us-east-1."
                "amazonaws.com/path/to/my/object",
            ),
            Case(
                ("https://s3-accesspoint-fips.dualstack.amazonaws.com",
                 "ap-south-1a"),
                "https://my-bucket.s3-accesspoint-fips.dualstack.ap-south-1a."
                "amazonaws.com/path/to/my/object",
            ),
            Case(
                ("https://s3-accesspoint-fips.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 None),
                "https://my-bucket.s3-accesspoint-fips.dualstack.us-gov-east-1."
                "amazonaws.com/path/to/my/object",
            ),
            Case(
                ("https://s3-accesspoint-fips.dualstack.us-gov-east-1."
                 "amazonaws.com",
                 "cn-northwest-1"),
                "https://my-bucket.s3-accesspoint-fips.dualstack."
                "cn-northwest-1.amazonaws.com/path/to/my/object",
            ),
            ###
            Case(
                ("https://my-load-balancer-1234567890.us-west-2.elb."
                 "amazonaws.com", "us-west-2"),
                "https://my-load-balancer-1234567890.us-west-2.elb."
                "amazonaws.com/my-bucket/path/to/my/object",
            ),
        ]

        for case in cases:
            base_url = BaseURL(*case.args)
            url = urlunsplit(base_url.build(
                "GET", base_url.region or "us-east-1",
                bucket_name="my-bucket", object_name="path/to/my/object"))
            self.assertEqual(str(url), case.result)



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage, (C)
# 2020 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Request/response of PutBucketReplication and GetBucketReplication APIs."""

from __future__ import absolute_import, annotations

from abc import ABCMeta
from typing import Type, TypeVar, cast
from xml.etree import ElementTree as ET

from .commonconfig import DISABLED, BaseRule, Filter, check_status
from .xml import Element, SubElement, find, findall, findtext

A = TypeVar("A", bound="Status")


class Status:
    """Status."""
    __metaclass__ = ABCMeta

    def __init__(self, status: str):
        check_status(status)
        self._status = status

    @property
    def status(self) -> str:
        """Get status."""
        return self._status

    @classmethod
    def fromxml(cls: Type[A], element: ET.Element) -> A:
        """Create new object with values from XML element."""
        element = cast(ET.Element, find(element, cls.__name__, True))
        status = cast(str, findtext(element, "Status", True))
        return cls(status)

    def toxml(self, element: ET.Element | None) -> ET.Element:
        """Convert to XML."""
        if element is None:
            raise ValueError("element must be provided")
        element = SubElement(element, self.__class__.__name__)
        SubElement(element, "Status", self._status)
        return element


class SseKmsEncryptedObjects(Status):
    """SSE KMS encrypted objects."""


B = TypeVar("B", bound="SourceSelectionCriteria")


class SourceSelectionCriteria:
    """Source selection criteria."""

    def __init__(
            self,
            sse_kms_encrypted_objects: SseKmsEncryptedObjects | None = None,
    ):
        self._sse_kms_encrypted_objects = sse_kms_encrypted_objects

    @property
    def sse_kms_encrypted_objects(self) -> SseKmsEncryptedObjects | None:
        """Get SSE KMS encrypted objects."""
        return self._sse_kms_encrypted_objects

    @classmethod
    def fromxml(cls: Type[B], element: ET.Element) -> B:
        """Create new object with values from XML element."""
        element = cast(
            ET.Element,
            find(element, "SourceSelectionCriteria", True),
        )
        return cls(
            None if find(element, "SseKmsEncryptedObjects") is None
            else SseKmsEncryptedObjects.fromxml(element)
        )

    def toxml(self, element: ET.Element | None) -> ET.Element:
        """Convert to XML."""
        if element is None:
            raise ValueError("element must be provided")
        element = SubElement(element, "SourceSelectionCriteria")
        if self._sse_kms_encrypted_objects:
            self._sse_kms_encrypted_objects.toxml(element)
        return element


class ExistingObjectReplication(Status):
    """Existing object replication."""


class DeleteMarkerReplication(Status):
    """Delete marker replication."""

    def __init__(self, status=DISABLED):
        super().__init__(status)


C = TypeVar("C", bound="ReplicationTimeValue")


class ReplicationTimeValue:
    """Replication time value."""
    __metaclass__ = ABCMeta

    def __init__(self, minutes: None | int = 15):
        self._minutes = minutes

    @property
    def minutes(self) -> int | None:
        """Get minutes."""
        return self._minutes

    @classmethod
    def fromxml(cls: Type[C], element: ET.Element) -> C:
        """Create new object with values from XML element."""
        element = cast(ET.Element, find(element, cls.__name__, True))
        minutes = findtext(element, "Minutes")
        return cls(int(minutes) if minutes else None)

    def toxml(self, element: ET.Element | None) -> ET.Element:
        """Convert to XML."""
        if element is None:
            raise ValueError("element must be provided")
        element = SubElement(element, self.__class__.__name__)
        if self._minutes is not None:
            SubElement(element, "Minutes", str(self._minutes))
        return element


class Time(ReplicationTimeValue):
    """Time."""


D = TypeVar("D", bound="ReplicationTime")


class ReplicationTime:
    """Replication time."""

    def __init__(self, time: Time, status: str):
        if not time:
            raise ValueError("time must be provided")
        check_status(status)
        self._time = time
        self._status = status

    @property
    def time(self) -> Time:
        """Get time value."""
        return self._time

    @property
    def status(self) -> str:
        """Get status."""
        return self._status

    @classmethod
    def fromxml(cls: Type[D], element: ET.Element) -> D:
        """Create new object with values from XML element."""
        element = cast(ET.Element, find(element, "ReplicationTime", True))
        time = Time.fromxml(element)
        status = cast(str, findtext(element, "Status", True))
        return cls(time, status)

    def toxml(self, element: ET.Element | None) -> ET.Element:
        """Convert to XML."""
        if element is None:
            raise ValueError("element must be provided")
        element = SubElement(element, "ReplicationTime")
        self._time.toxml(element)
        SubElement(element, "Status", self._status)
        return element


class EventThreshold(ReplicationTimeValue):
    """Event threshold."""


E = TypeVar("E", bound="Metrics")


class Metrics:
    """Metrics."""

    def __init__(self, event_threshold: EventThreshold, status: str):
        if not event_threshold:
            raise ValueError("event threshold must be provided")
        check_status(status)
        self._event_threshold = event_threshold
        self._status = status

    @property
    def event_threshold(self) -> EventThreshold:
        """Get event threshold."""
        return self._event_threshold

    @property
    def status(self) -> str:
        """Get status."""
        return self._status

    @classmethod
    def fromxml(cls: Type[E], element: ET.Element) -> E:
        """Create new object with values from XML element."""
        element = cast(ET.Element, find(element, "Metrics", True))
        event_threshold = EventThreshold.fromxml(element)
        status = cast(str, findtext(element, "Status", True))
        return cls(event_threshold, status)

    def toxml(self, element: ET.Element | None) -> ET.Element:
        """Convert to XML."""
        if element is None:
            raise ValueError("element must be provided")
        element = SubElement(element, "Metrics")
        self._event_threshold.toxml(element)
        SubElement(element, "Status", self._status)
        return element


F = TypeVar("F", bound="EncryptionConfig")


class EncryptionConfig:
    """Encryption configuration."""

    def __init__(self, replica_kms_key_id: str | None = None):
        self._replica_kms_key_id = replica_kms_key_id

    @property
    def replica_kms_key_id(self) -> str | None:
        """Get replica KMS key ID."""
        return self._replica_kms_key_id

    @classmethod
    def fromxml(cls: Type[F], element: ET.Element) -> F:
        """Create new object with values from XML element."""
        element = cast(
            ET.Element,
            find(element, "EncryptionConfiguration", True),
        )
        return cls(findtext(element, "ReplicaKmsKeyID"))

    def toxml(self, element: ET.Element | None) -> ET.Element:
        """Convert to XML."""
        if element is None:
            raise ValueError("element must be provided")
        element = SubElement(element, "EncryptionConfiguration")
        SubElement(element, "ReplicaKmsKeyID", self._replica_kms_key_id)
        return element


G = TypeVar("G", bound="AccessControlTranslation")


class AccessControlTranslation:
    """Access control translation."""

    def __init__(self, owner: str = "Destination"):
        if not owner:
            raise ValueError("owner must be provided")
        self._owner = owner

    @property
    def owner(self) -> str:
        """Get owner."""
        return self._owner

    @classmethod
    def fromxml(cls: Type[G], element: ET.Element) -> G:
        """Create new object with values from XML element."""
        element = cast(
            ET.Element, find(element, "AccessControlTranslation", True),
        )
        owner = cast(str, findtext(element, "Owner", True))
        return cls(owner)

    def toxml(self, element: ET.Element | None) -> ET.Element:
        """Convert to XML."""
        if element is None:
            raise ValueError("element must be provided")
        element = SubElement(element, "AccessControlTranslation")
        SubElement(element, "Owner", self._owner)
        return element


H = TypeVar("H", bound="Destination")


class Destination:
    """Replication destination."""

    def __init__(
            self,
            bucket_arn: str,
            access_control_translation: AccessControlTranslation | None = None,
            account: str | None = None,
            encryption_config: EncryptionConfig | None = None,
            metrics: Metrics | None = None,
            replication_time: ReplicationTime | None = None,
            storage_class: str | None = None,
    ):
        if not bucket_arn:
            raise ValueError("bucket ARN must be provided")
        self._bucket_arn = bucket_arn
        self._access_control_translation = access_control_translation
        self._account = account
        self._encryption_config = encryption_config
        self._metrics = metrics
        self._replication_time = replication_time
        self._storage_class = storage_class

    @property
    def bucket_arn(self) -> str:
        """Get bucket ARN."""
        return self._bucket_arn

    @property
    def access_control_translation(self) -> AccessControlTranslation | None:
        """Get access control translation. """
        return self._access_control_translation

    @property
    def account(self) -> str | None:
        """Get account."""
        return self._account

    @property
    def encryption_config(self) -> EncryptionConfig | None:
        """Get encryption configuration."""
        return self._encryption_config

    @property
    def metrics(self) -> Metrics | None:
        """Get metrics."""
        return self._metrics

    @property
    def replication_time(self) -> ReplicationTime | None:
        """Get replication time."""
        return self._replication_time

    @property
    def storage_class(self) -> str | None:
        """Get storage class."""
        return self._storage_class

    @classmethod
    def fromxml(cls: Type[H], element: ET.Element) -> H:
        """Create new object with values from XML element."""
        element = cast(ET.Element, find(element, "Destination", True))
        access_control_translation = (
            None if find(element, "AccessControlTranslation") is None
            else AccessControlTranslation.fromxml(element)
        )
        account = findtext(element, "Account")
        bucket_arn = cast(str, findtext(element, "Bucket", True))
        encryption_config = (
            None if find(element, "EncryptionConfiguration") is None
            else EncryptionConfig.fromxml(element)
        )
        metrics = (
            None if find(element, "Metrics") is None
            else Metrics.fromxml(element)
        )
        replication_time = (
            None if find(element, "ReplicationTime") is None
            else ReplicationTime.fromxml(element)
        )
        storage_class = findtext(element, "StorageClass")
        return cls(bucket_arn, access_control_translation, account,
                   encryption_config, metrics, replication_time, storage_class)

    def toxml(self, element: ET.Element | None) -> ET.Element:
        """Convert to XML."""
        if element is None:
            raise ValueError("element must be provided")
        element = SubElement(element, "Destination")
        if self._access_control_translation:
            self._access_control_translation.toxml(element)
        if self._account is not None:
            SubElement(element, "Account", self._account)
        SubElement(element, "Bucket", self._bucket_arn)
        if self._encryption_config:
            self._encryption_config.toxml(element)
        if self._metrics:
            self._metrics.toxml(element)
        if self._replication_time:
            self._replication_time.toxml(element)
        if self._storage_class:
            SubElement(element, "StorageClass", self._storage_class)
        return element


I = TypeVar("I", bound="Rule")


class Rule(BaseRule):
    """Replication rule. """

    def __init__(
            self,
            destination: Destination,
            status: str,
            delete_marker_replication: DeleteMarkerReplication | None = None,
            existing_object_replication:
                ExistingObjectReplication | None = None,
            rule_filter: Filter | None = None,
            rule_id: str | None = None,
            prefix: str | None = None,
            priority: int | None = None,
            source_selection_criteria: SourceSelectionCriteria | None = None,
    ):
        if not destination:
            raise ValueError("destination must be provided")

        check_status(status)

        super().__init__(rule_filter, rule_id)

        self._destination = destination
        self._status = status
        if rule_filter and not delete_marker_replication:
            delete_marker_replication = DeleteMarkerReplication()
        self._delete_marker_replication = delete_marker_replication
        self._existing_object_replication = existing_object_replication
        self._prefix = prefix
        self._priority = priority
        self._source_selection_criteria = source_selection_criteria

    @property
    def destination(self) -> Destination:
        """Get destination."""
        return self._destination

    @property
    def status(self) -> str:
        """Get status."""
        return self._status

    @property
    def delete_marker_replication(self) -> DeleteMarkerReplication | None:
        """Get delete marker replication."""
        return self._delete_marker_replication

    @property
    def existing_object_replication(self) -> ExistingObjectReplication | None:
        """Get existing object replication."""
        return self._existing_object_replication

    @property
    def prefix(self) -> str | None:
        """Get prefix."""
        return self._prefix

    @property
    def priority(self) -> int | None:
        """Get priority."""
        return self._priority

    @property
    def source_selection_criteria(self) -> SourceSelectionCriteria | None:
        """Get source selection criteria."""
        return self._source_selection_criteria

    @classmethod
    def fromxml(cls: Type[I], element: ET.Element) -> I:
        """Create new object with values from XML element."""
        delete_marker_replication = (
            None if find(element, "DeleteMarkerReplication") is None
            else DeleteMarkerReplication.fromxml(element)
        )
        destination = Destination.fromxml(element)
        existing_object_replication = (
            None if find(element, "ExistingObjectReplication") is None
            else ExistingObjectReplication.fromxml(element)
        )
        rule_filter, rule_id = cls.parsexml(element)
        prefix = findtext(element, "Prefix")
        priority = findtext(element, "Priority")
        source_selection_criteria = (
            None if find(element, "SourceSelectionCriteria") is None
            else SourceSelectionCriteria.fromxml(element)
        )
        status = cast(str, findtext(element, "Status", True))

        return cls(
            destination,
            status,
            delete_marker_replication,
            existing_object_replication,
            rule_filter,
            rule_id,
            prefix,
            int(priority) if priority else None,
            source_selection_criteria,
        )

    def toxml(self, element: ET.Element | None) -> ET.Element:
        """Convert to XML."""
        if element is None:
            raise ValueError("element must be provided")
        element = SubElement(element, "Rule")
        if self._delete_marker_replication:
            self._delete_marker_replication.toxml(element)
        self._destination.toxml(element)
        if self._existing_object_replication:
            self._existing_object_replication.toxml(element)
        super().toxml(element)
        if self._prefix is not None:
            SubElement(element, "Prefix", self._prefix)
        if self._priority is not None:
            SubElement(element, "Priority", str(self._priority))
        if self._source_selection_criteria:
            self._source_selection_criteria.toxml(element)
        SubElement(element, "Status", self._status)
        return element


J = TypeVar("J", bound="ReplicationConfig")


class ReplicationConfig:
    """Replication configuration."""

    def __init__(self, role: str, rules: list[Rule]):
        if not rules:
            raise ValueError("rules must be provided")
        if len(rules) > 1000:
            raise ValueError("more than 1000 rules are not supported")
        self._role = role
        self._rules = rules

    @property
    def role(self) -> str:
        """Get role."""
        return self._role

    @property
    def rules(self) -> list[Rule]:
        """Get rules."""
        return self._rules

    @classmethod
    def fromxml(cls: Type[J], element: ET.Element) -> J:
        """Create new object with values from XML element."""
        role = cast(str, findtext(element, "Role", True))
        elements = findall(element, "Rule")
        rules = []
        for tag in elements:
            rules.append(Rule.fromxml(tag))
        return cls(role, rules)

    def toxml(self, element: ET.Element | None) -> ET.Element:
        """Convert to XML."""
        element = Element("ReplicationConfiguration")
        SubElement(element, "Role", self._role)
        for rule in self._rules:
            rule.toxml(element)
        return element



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage,
# (C) 2020 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json

import urllib3

from minio import Minio
from minio.credentials import WebIdentityProvider


def get_jwt(client_id, client_secret, idp_client_id, idp_endpoint):
    res = urllib3.PoolManager().request(
        "POST",
        idp_endpoint,
        fields={
            "username": client_id,
            "password": client_secret,
            "grant_type": "password",
            "client_id": idp_client_id,
        },
    )

    return json.loads(res.data.encode())


# IDP endpoint.
idp_endpoint = (
    "https://IDP-HOST:IDP-PORT/auth/realms/master"
    "/protocol/openid-connect/token"
)

# Client-ID to fetch JWT.
client_id = "USER-ID"

# Client secret to fetch JWT.
client_secret = "PASSWORD"

# Client-ID of MinIO service on IDP.
idp_client_id = "MINIO-CLIENT-ID"

# STS endpoint usually point to MinIO server.
sts_endpoint = "http://STS-HOST:STS-PORT/"

# Role ARN if available.
role_arn = "ROLE-ARN"

# Role session name if available.
role_session_name = "ROLE-SESSION-NAME"

provider = WebIdentityProvider(
    lambda: get_jwt(client_id, client_secret, idp_client_id, idp_endpoint),
    sts_endpoint,
    role_arn=role_arn,
    role_session_name=role_session_name,
)

client = Minio("MINIO-HOST:MINIO-PORT", credentials=provider)

# Get information of an object.
stat = client.stat_object("my-bucket", "my-object")
print(stat)



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage,
# (C) 2015, 2016, 2017 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import TestCase
from urllib.parse import urlunsplit

from minio import Minio
from minio import __version__ as minio_version
from minio.api import _DEFAULT_USER_AGENT
from minio.helpers import BaseURL, check_bucket_name


class ValidBucketName(TestCase):
    def test_bucket_name(self):
        self.assertRaises(ValueError, check_bucket_name, 'bucketName=', False)

    def test_bucket_name_invalid_characters(self):
        self.assertRaises(ValueError, check_bucket_name, '$$$bcuket', False)

    def test_bucket_name_length(self):
        self.assertRaises(ValueError, check_bucket_name, 'dd', False)

    def test_bucket_name_periods(self):
        self.assertRaises(ValueError, check_bucket_name, 'dd..mybucket', False)

    def test_bucket_name_begins_period(self):
        self.assertRaises(ValueError, check_bucket_name, '.ddmybucket', False)


class GetURLTests(TestCase):
    def test_url_build(self):
        url = BaseURL('http://localhost:9000', None)
        self.assertEqual(
            urlunsplit(url.build("GET", None, bucket_name='bucket-name')),
            'http://localhost:9000/bucket-name',
        )
        self.assertEqual(
            urlunsplit(
                url.build("GET", None, bucket_name='bucket-name',
                          object_name='objectName'),
            ),
            'http://localhost:9000/bucket-name/objectName',
        )
        self.assertEqual(
            urlunsplit(
                url.build("GET", 'us-east-1', bucket_name='bucket-name',
                          object_name='objectName',
                          query_params={'foo': 'bar'}),
            ),
            'http://localhost:9000/bucket-name/objectName?foo=bar',
        )
        self.assertEqual(
            urlunsplit(
                url.build("GET", 'us-east-1', bucket_name='bucket-name',
                          object_name='objectName',
                          query_params={'foo': 'bar', 'b': 'c', 'a': 'b'}),
            ),
            'http://localhost:9000/bucket-name/objectName?a=b&b=c&foo=bar',
        )
        self.assertEqual(
            urlunsplit(
                url.build("GET", 'us-east-1', bucket_name='bucket-name',
                          object_name='path/to/objectName/'),
            ),
            'http://localhost:9000/bucket-name/path/to/objectName/',
        )

        # S3 urls.
        url = BaseURL('https://s3.amazonaws.com', None)
        self.assertEqual(
            urlunsplit(url.build("GET", "us-east-1")),
            'https://s3.us-east-1.amazonaws.com/',
        )
        self.assertEqual(
            urlunsplit(
                url.build("GET", "eu-west-1", bucket_name='my.bucket.name'),
            ),
            'https://s3.eu-west-1.amazonaws.com/my.bucket.name',
        )
        self.assertEqual(
            urlunsplit(
                url.build("GET", 'us-west-2', bucket_name='bucket-name',
                          object_name='objectName'),
            ),
            'https://bucket-name.s3.us-west-2.amazonaws.com/objectName',
        )
        self.assertEqual(
            urlunsplit(
                url.build("GET", "us-east-1", bucket_name='bucket-name',
                          object_name='objectName',
                          query_params={'versionId': 'uuid'}),
            ),
            "https://bucket-name.s3.us-east-1.amazonaws.com"
            "/objectName?versionId=uuid",
        )

    def test_minio_requires_string(self):
        self.assertRaises(TypeError, Minio, 10)

    def test_minio_requires_hostname(self):
        self.assertRaises(ValueError, Minio, 'http://')


class UserAgentTests(TestCase):
    def test_default_user_agent(self):
        client = Minio('localhost')
        self.assertEqual(client._user_agent, _DEFAULT_USER_AGENT)

    def test_set_app_info(self):
        client = Minio('localhost')
        expected_user_agent = _DEFAULT_USER_AGENT + ' hello/' + minio_version
        client.set_app_info('hello', minio_version)
        self.assertEqual(client._user_agent, expected_user_agent)

    def test_set_app_info_requires_non_empty_name(self):
        client = Minio('localhost:9000')
        self.assertRaises(ValueError, client.set_app_info, '', minio_version)

    def test_set_app_info_requires_non_empty_version(self):
        client = Minio('localhost:9000')
        self.assertRaises(ValueError, client.set_app_info, 'hello', '')


class GetRegionTests(TestCase):
    def test_region_none(self):
        region = BaseURL('http://localhost', None).region
        self.assertIsNone(region)

    def test_region_us_west(self):
        region = BaseURL('https://s3-us-west-1.amazonaws.com', None).region
        self.assertEqual(region, "")

    def test_region_with_dot(self):
        region = BaseURL('https://s3.us-west-1.amazonaws.com', None).region
        self.assertEqual(region, 'us-west-1')

    def test_region_with_dualstack(self):
        region = BaseURL(
            'https://s3.dualstack.us-west-1.amazonaws.com', None,
        ).region
        self.assertEqual(region, 'us-west-1')

    def test_region_us_east(self):
        region = BaseURL('http://s3.amazonaws.com', None).region
        self.assertEqual(region, "")

    def test_invalid_value(self):
        self.assertRaises(ValueError, BaseURL, None, None)



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage,
# (C) 2020 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import TestCase

from minio import xml
from minio.notificationconfig import (NotificationConfig, PrefixFilterRule,
                                      QueueConfig)


class NotificationConfigTest(TestCase):
    def test_config(self):
        config = NotificationConfig(
            queue_config_list=[
                QueueConfig(
                    "QUEUE-ARN-OF-THIS-BUCKET",
                    ['s3:ObjectCreated:*'],
                    config_id="1",
                    prefix_filter_rule=PrefixFilterRule("abc"),
                ),
            ],
        )
        xml.marshal(config)

        config = xml.unmarshal(
            NotificationConfig,
            """<NotificationConfiguration>
  <CloudFunctionConfiguration>
    <Id>ObjectCreatedEvents</Id>
    <CloudFunction>arn:aws:lambda:us-west-2:35667example:function:CreateThumbnail</CloudFunction>
    <Event>s3:ObjectCreated:*</Event>
  </CloudFunctionConfiguration>
  <QueueConfiguration>
      <Id>1</Id>
      <Filter>
          <S3Key>
              <FilterRule>
                  <Name>prefix</Name>
                  <Value>images/</Value>
              </FilterRule>
              <FilterRule>
                  <Name>suffix</Name>
                  <Value>.jpg</Value>
              </FilterRule>
          </S3Key>
     </Filter>
     <Queue>arn:aws:sqs:us-west-2:444455556666:s3notificationqueue</Queue>
     <Event>s3:ObjectCreated:Put</Event>
  </QueueConfiguration>
</NotificationConfiguration>""",
        )
        xml.marshal(config)



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage,
# (C) 2015 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from io import BytesIO
from unittest import TestCase

from urllib3.response import HTTPResponse

from minio.crypto import decrypt, encrypt


class CryptoTest(TestCase):
    def test_correct(self):
        secret = "topsecret"
        plaintext = "Hello MinIO!"
        encrypted = encrypt(plaintext.encode(), secret)
        decrypted = decrypt(
            HTTPResponse(body=BytesIO(encrypted), preload_content=False),
            secret,
        ).decode()
        if hasattr(self, "assertEquals"):
            self.assertEquals(plaintext, decrypted)
        else:
            self.assertEqual(plaintext, decrypted)

    def test_wrong(self):
        secret = "topsecret"
        secret2 = "othersecret"
        plaintext = "Hello MinIO!"
        encrypted = encrypt(plaintext.encode(), secret)
        self.assertRaises(
            ValueError,
            decrypt,
            HTTPResponse(body=BytesIO(encrypted), preload_content=False),
            secret2,
        )



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage,
# (C) 2015 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import hmac
from datetime import datetime, timezone
from unittest import TestCase
from urllib.parse import urlsplit, urlunsplit

from minio import Minio
from minio.credentials import Credentials
from minio.helpers import queryencode, quote, sha256_hash
from minio.signer import (_get_authorization, _get_canonical_request_hash,
                          _get_scope, _get_signing_key, _get_string_to_sign,
                          presign_v4, sign_v4_s3)

empty_hash = 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
dt = datetime(2015, 6, 20, 1, 2, 3, 0, timezone.utc)


class CanonicalRequestTest(TestCase):
    def test_simple_request(self):
        url = urlsplit('http://localhost:9000/hello')
        expected_signed_headers = ['x-amz-content-sha256', 'x-amz-date']
        expected_request_array = ['PUT', '/hello', '',
                                  'x-amz-content-sha256:' +
                                  empty_hash, 'x-amz-date:dateString',
                                  '', ';'.join(expected_signed_headers),
                                  empty_hash]
        headers_to_sign = {'x-amz-date': 'dateString',
                           'x-amz-content-sha256': empty_hash}

        expected_request = sha256_hash('\n'.join(expected_request_array))
        actual_request = _get_canonical_request_hash(
            "PUT", url, headers_to_sign, empty_hash,
        )
        self.assertEqual(expected_request, actual_request[0])

    def test_request_with_query(self):
        url = urlsplit('http://localhost:9000/hello?c=d&e=f&a=b')
        expected_signed_headers = ['x-amz-content-sha256', 'x-amz-date']
        expected_request_array = ['PUT', '/hello', 'a=b&c=d&e=f',
                                  'x-amz-content-sha256:' + empty_hash,
                                  'x-amz-date:dateString',
                                  '', ';'.join(expected_signed_headers),
                                  empty_hash]

        expected_request = sha256_hash('\n'.join(expected_request_array))

        headers_to_sign = {'x-amz-date': 'dateString',
                           'x-amz-content-sha256': empty_hash}
        actual_request = _get_canonical_request_hash(
            "PUT", url, headers_to_sign, empty_hash,
        )
        self.assertEqual(expected_request, actual_request[0])


class StringToSignTest(TestCase):
    def test_signing_key(self):
        expected_signing_key_list = [
            'AWS4-HMAC-SHA256', '20150620T010203Z',
            '20150620/us-east-1/s3/aws4_request',
            'b93e86965c269a0dfef37a8bec231ef8acf8cdb101a64eb700a46c452c1ad233'
        ]

        actual_signing_key = _get_string_to_sign(
            dt, _get_scope(dt, 'us-east-1', "s3"),
            'b93e86965c269a0dfef37a8bec231ef8acf8cdb101a64eb700a46c452c1ad233')
        self.assertEqual('\n'.join(expected_signing_key_list),
                         actual_signing_key)


class SigningKeyTest(TestCase):
    def test_generate_signing_key(self):
        key1_string = 'AWS4' + 'S3CR3T'
        key1 = key1_string.encode('utf-8')
        key2 = hmac.new(key1, '20150620'.encode(
            'utf-8'), hashlib.sha256).digest()
        key3 = hmac.new(key2, 'region'.encode(
            'utf-8'), hashlib.sha256).digest()
        key4 = hmac.new(key3, 's3'.encode('utf-8'), hashlib.sha256).digest()
        expected_result = hmac.new(key4, 'aws4_request'.encode(
            'utf-8'), hashlib.sha256).digest()

        actual_result = _get_signing_key('S3CR3T', dt, 'region', "s3")
        self.assertEqual(expected_result, actual_result)


class AuthorizationHeaderTest(TestCase):
    def test_generate_authentication_header(self):
        expected_authorization_header = (
            'AWS4-HMAC-SHA256 Credential='
            'public_key/20150620/region/s3/aws4_request, '
            'SignedHeaders=host;X-Amz-Content-Sha256;X-Amz-Date, '
            'Signature=signed_request'
        )
        actual_authorization_header = _get_authorization(
            'public_key', _get_scope(dt, 'region', "s3"),
            'host;X-Amz-Content-Sha256;X-Amz-Date', 'signed_request')
        self.assertEqual(expected_authorization_header,
                         actual_authorization_header)


class PresignURLTest(TestCase):
    def test_presigned_versioned_id(self):
        credentials = Credentials("minio", "minio123")
        url = presign_v4('GET', urlsplit('http://localhost:9000/bucket-name/objectName?versionId=uuid'),
                         'us-east-1', credentials, dt, 604800)

        self.assertEqual(urlunsplit(url), 'http://localhost:9000/bucket-name/objectName?versionId=uuid&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=minio%2F20150620%2Fus-east-1%2Fs3%2Faws4_request&X-Amz-Date=20150620T010203Z&X-Amz-Expires=604800&X-Amz-SignedHeaders=host&X-Amz-Signature=3ce13e2ca929fafa20581a05730e4e9435f2a5e20ec7c5a082d175692fb0a663')


class SignV4Test(TestCase):
    def test_signv4(self):
        client = Minio("localhost:9000", access_key="minio",
                       secret_key="minio123", secure=False)
        creds = client._provider.retrieve()
        headers = {
            'Host': 'localhost:9000',
            'x-amz-content-sha256':
            'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855',
            'x-amz-date': '20150620T010203Z',
        }
        url = client._base_url.build(
            "PUT",
            "us-east-1",
            bucket_name="testbucket",
            object_name="~testobject",
            query_params={"partID": "1", "uploadID": "~abcd"},
        )
        headers = sign_v4_s3(
            "PUT",
            url,
            "us-east-1",
            headers,
            creds,
            "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
            dt,
        )
        self.assertEqual(headers['Authorization'],
                         'AWS4-HMAC-SHA256 Credential='
                         'minio/20150620/us-east-1/s3/aws4_request, '
                         'SignedHeaders=host;x-amz-content-sha256;x-amz-date, '
                         'Signature='
                         'a2f4546f647981732bd90dfa5a7599c44dca92f44bea48ecc7565df06032c25b')


class UnicodeEncodeTest(TestCase):
    def test_unicode_quote(self):
        self.assertEqual(quote('/test/123/汉字'), '/test/123/%E6%B1%89%E5%AD%97')

    def test_unicode_queryencode(self):
        self.assertEqual(queryencode('/test/123/汉字'),
                         '%2Ftest%2F123%2F%E6%B1%89%E5%AD%97')

    def test_unicode_quote_u(self):
        self.assertEqual(quote(u'/test/123/汉字'),
                         '/test/123/%E6%B1%89%E5%AD%97')

    def test_unicode_queryencode_u(self):
        self.assertEqual(queryencode(u'/test/123/汉字'),
                         '%2Ftest%2F123%2F%E6%B1%89%E5%AD%97')

    def test_unicode_quote_b(self):
        self.assertEqual(quote(b'/test/123/\xe6\xb1\x89\xe5\xad\x97'),
                         '/test/123/%E6%B1%89%E5%AD%97')

    def test_unicode_queryencode_b(self):
        self.assertEqual(queryencode(b'/test/123/\xe6\xb1\x89\xe5\xad\x97'),
                         '%2Ftest%2F123%2F%E6%B1%89%E5%AD%97')
